from .Schema import Schema
from .SchemaAnything import SchemaAnything
from .SchemaType import SchemaType
from .SchemaPrimitive import SchemaPrimitive
from .SchemaNone import SchemaNone
from .SchemaConst import SchemaConst




from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

from jsonschema.validators import validator_for


class Schema(ABC):

    @abstractmethod
    def __call__(
            self,
            value: Any,
            dom_info: Tuple = None
    ) -> Any:
        return value

    @property
    @abstractmethod
    def schema(self) -> Dict[str, Any]:
        return {}

    def is_valid(self, value: Any) -> bool:
        return validator_for(self.schema)(self.schema).is_valid(value)



from .__version__ import __version__
from . import mixins
from .exceptions import ValidationError
from .dom import document, parent, key, schema
from .dom import DOMInfo as DOMInfo
from .dom import DOMElement as Element
from .base_schema import Schema, SchemaAnything, SchemaConst
from .object_schema import SchemaArray, SchemaDict
from .user_objects import UserProperty, UserObject
from .user_object_mixins import ReadsJSON, ReadsYAML



from typing import Optional, Dict, Type, Any, Union

from ..dom import DOMInfo
from ..base_schema import SchemaType
from ..base_schema import Schema


class SchemaObject(SchemaType):

    type_name: str = "object"
    properties: Optional[Dict[str, Schema]] = None
    additional_properties: Union[bool, Schema] = False

    def __init__(
            self,
            properties: Optional[Dict[str, Schema]] = None,
            additional_properties: Union[bool, Schema] = False,
            object_type: Type = dict
    ) -> None:
        self.properties = properties or {}
        self.additional_properties = additional_properties
        self.object_type = object_type

    def __call__(
            self,
            value: Any,
            dom_info: DOMInfo = None
    ) -> Any:
        return self.object_type(value, dom_info)

    @property
    def schema(self) -> Dict[str, Any]:
        return {
            **super().schema,
            'properties': {
                k: v.schema
                for k, v in self.properties.items()
            },
            "additionalProperties": (
                self.additional_properties.schema
                if isinstance(self.additional_properties, Schema)
                else self.additional_properties
            )
        }





from typing import Any, Union, Type, Dict, Iterable

from ..dom import DOMInfo, DOMList
from ..base_schema import Schema

from .resolve_arg_to_type import resolve_arg_to_schema


class SchemaArray(Schema):

    items: Schema = None

    def __init__(
            self,
            items: Union[Type, Schema]
    ) -> None:
        self.items = resolve_arg_to_schema(items)

    def __call__(
            self,
            value: Iterable,
            dom_info: DOMInfo = None
    ) -> Any:
        return DOMList(
            value,
            dom_info,
            _item_type=self.items
        )

    @property
    def schema(self) -> Dict[str, Any]:
        return {
            "array": {
                "items": self.items
            }
        }



from typing import Any, Dict, Tuple

from .Schema import Schema


class SchemaConst(Schema):

    value: str = None

    def __init__(
            self,
            value: str
    ) -> None:
        self.value = value

    def __call__(
            self,
            value: str,
            dom_info: Tuple = None
    ) -> Any:
        if value != self.value:
            raise ValueError(f"Value can only be '{self.value}'.")
        return self.value

    @property
    def schema(self) -> Dict[str, Any]:
        return {
            "const": self.value
        }



from __future__ import annotations

from abc import ABC, abstractmethod

from typing import Any, Dict

from .Schema import Schema


class SchemaType(Schema, ABC):

    @property
    @abstractmethod
    def type_name(self) -> str:
        pass

    @property
    def schema(self) -> Dict[str, Any]:
        return {"type": self.type_name}





from __future__ import annotations

from abc import ABC

from typing import Type, Optional, Any, Dict

import inspect


class RegistersSubclasses(ABC):
    """
    A mixin class that allows subclasses to be registered in the
    main class by a registered name, using the `register_as`
    parameter in the subclass declaration.
    """

    registered_name = None
    _registration_namespace = None
    _registered_subclasses = {}

    def __init_subclass__(
        cls,
        register_as: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init_subclass__(**kwargs)

        cls._registration_namespace = cls._get_registration_namespace()

        if register_as:
            key = (cls.registration_namespace(), str(register_as))
            if (
                key in cls._registered_subclasses
                and cls._registered_subclasses[key] is not cls
            ):
                raise Exception(
                    f"""
                    Cannot register {cls} as {register_as}:
                    Already used by {cls._registered_subclasses[key]}.
                    """)
            cls._registered_subclasses[key] = cls
            cls.registered_name = str(register_as)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def registration_namespace(cls) -> str:
        """
        :return: The registration namespace for this class.
        """
        try:
            return cls._registration_namespace
        except AttributeError:
            raise AttributeError(
                "Cannot find attribute '_registration_namespace'. "
                "Has another class overridden __init_subclass__ "
                "without calling super().__init_subclass__?"
            )

    @classmethod
    def registered_subclasses(cls) -> Dict[str, RegisteredSubclass]:
        """
        Return all of the registered subclasses in this class's namespace.

        :return: A dictionary of subclasses, indexed by registered name.
        """
        return {
            name: subclass
            for namespace, name, subclass in [
                (*key, subclass)
                for key, subclass in cls._registered_subclasses.items()
            ]
            if namespace == cls.registration_namespace()
            and issubclass(subclass, cls)
            and subclass is not cls
        }

    @classmethod
    def is_base_class(cls) -> bool:
        """
        Return True if this class is the base class for the namespace,
        i.e. the class that first declared RegisteredSubclasses as
        one of its bases.
        """
        return cls._get_class_namespace(cls) == cls.registration_namespace()

    @classmethod
    def registered_subclass(cls, name: str) -> RegisteredSubclass:
        """
        Return a registered subclass by name.

        :param name: The registered name of the subclass, as defined in
                     `register_as` when the class was declared.
        :return:     A subclass of the base class for the namespace.
        """
        key = (cls.registration_namespace(), name)
        if key not in cls._registered_subclasses:
            raise KeyError(
                f"Unknown registered subclass key: {name}")
        return cls._registered_subclasses[key]

    @classmethod
    def registered_subclass_instance(
        cls,
        name: str,
        *args: Any,
        **kwargs: Any
    ) -> RegistersSubclasses:
        """
        Create a new instance of a registered subclass by name.

        :param name:   The registered name of the subclass, as defined in
                       `register_as` when the class was declared.
        :param args:   Positional arguments to pass to the subclass.
        :param kwargs: Keyword arguments to pass to the subclass.
        :return:
        """
        return cls.registered_subclass(name)(*args, **kwargs)

    @classmethod
    def _get_registration_namespace(cls) -> str:
        return [
            cls._get_class_namespace(ancestor)
            for ancestor in inspect.getmro(cls)
            if RegistersSubclasses in ancestor.__bases__
        ][0]

    @staticmethod
    def _get_class_namespace(cls: type) -> str:
        return f"{cls.__module__}.{cls.__name__}"


RegisteredSubclass = Type[RegistersSubclasses]



from __future__ import annotations

from typing import Generic, TypeVar, Optional, Any

from collections.abc import Mapping

from ..base_schema import Schema, SchemaAnything
from .DOMObject import DOMObject
from . import DOMInfo
from .DOMProperties import DOMProperties

T_co = TypeVar('T_co')


class DOMDict(DOMObject, Generic[T_co]):

    _additional_properties: Schema = None

    def __init__(
            self,
            value: Optional[Mapping[str, Any]] = None,
            json_dom_info: Optional[DOMInfo] = None,
            _item_type: Optional[Schema] = None
    ) -> None:
        self.__json_schema_properties__ = DOMProperties(
            additional_properties=(_item_type or SchemaAnything())
        )
        super().__init__(
            value or {},
            json_dom_info
        )

    def __getitem__(self, key: str) -> T_co:
        return super().__getitem__(key)



from __future__ import annotations

from typing import Any, Iterator, Dict

from collections.abc import Mapping, MutableMapping

from copy import deepcopy

from wysdom.mixins import RegistersSubclasses

from ..exceptions import ValidationError
from ..base_schema import SchemaAnything

from .DOMElement import DOMElement
from . import DOMInfo
from .DOMProperties import DOMProperties
from .functions import document


class DOMObject(DOMElement, MutableMapping, RegistersSubclasses):

    __json_schema_properties__: DOMProperties = None
    __json_element_data__: Dict[str, DOMElement] = None

    def __init__(
            self,
            value: Mapping[str, Any] = None,
            json_dom_info: DOMInfo = None
    ) -> None:
        if value and not isinstance(value, Mapping):
            raise ValidationError(
                f"Cannot validate input. Object is not a mapping: {value}"
            )
        super().__init__(None, json_dom_info)
        self.__json_element_data__ = {}
        try:
            for key, value in value.items():
                self[key] = value
        except KeyError as e:
            raise ValidationError(str(e))

    def __getitem__(self, key: str) -> DOMElement:
        return self.__json_element_data__[key]

    def __setitem__(self, key: str, value: DOMElement) -> None:
        item_class = self.__json_schema_properties__.properties.get(
            key, self.__json_schema_properties__.additional_properties)
        if item_class is True:
            item_class = SchemaAnything()
        if not item_class:
            raise KeyError(
                f"No property named '{key}' exists, and "
                "additional properties are not allowed."
            )
        self.__json_element_data__[key] = item_class(
            value,
            DOMInfo(
                document=document(self),
                parent=self,
                element_key=key
            )
        )

    def __delitem__(self, key: str) -> None:
        del self.__json_element_data__[key]

    def __len__(self) -> int:
        return len(self.__json_element_data__)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__json_element_data__)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.__json_element_data__)})"

    def __str__(self):
        return str(self.__json_element_data__)

    def walk_elements(self) -> Iterator[DOMInfo]:
        for key, value in self.items():
            if isinstance(value, DOMElement):
                yield from value.walk_elements()
            else:
                yield DOMInfo(value, document(self), self, key)

    def to_builtin(self) -> Dict[str, Any]:
        """
        Returns the contents of this DOM object as a Python builtin.

        :return: A Python dict containing this object's data
        """
        return {
            k: (
                v.to_builtin()
                if isinstance(v, DOMElement)
                else v
            )
            for k, v in self.items()
        }

    def __copy__(self) -> DOMObject:
        cls = self.__class__
        return cls(dict(self))

    def __deepcopy__(self, memo: Dict[int, DOMElement]) -> DOMObject:
        cls = self.__class__
        result = cls(deepcopy(dict(self), memo))
        memo[id(self)] = result
        return result



from __future__ import annotations

from typing import Type, Union

import inspect

from ..base_schema import Schema
from ..dom import DOMElement

from ..base_schema import SchemaPrimitive


def resolve_arg_to_schema(
        arg: Union[Type, Schema]
) -> Schema:
    if inspect.isclass(arg):
        if issubclass(arg, DOMElement):
            return arg.__json_schema__()
        else:
            return SchemaPrimitive(arg)
    elif isinstance(arg, Schema):
        return arg
    else:
        raise TypeError(f"Unexpected object type: {type(arg)}")



from typing import Type, Any, Union, Optional, Callable

from ..base_schema import Schema
from ..object_schema import resolve_arg_to_schema
from ..dom import DOMObject


class UserProperty(object):
    """
    A data descriptor for creating attributes in user-defined subclasses
    of `UserObject` which are mapped to keys in the underlying
    data object and to the `properties` key in the object's JSON schema.

    :param property_type:    The data type or schema for this property. Must
                             be one of:
                             A primitive Python type (str, int, bool, float)
                             A subclass of `UserObject`
                             An instance of `JSONSchema`

    :param name:             The name of this property in the underlying
                             data object. If not provided, this defaults to
                             the name of the attribute on the `UserObject`
                             instance that owns the property.

    :param default:          A static value which provides a default value
                             for this property. Cannot be set in conjunction
                             with `default_function`.

    :param default_function: A function which provides a default value
                             for this property. The function must have a
                             single positional argument, `self`, which is
                             passed the `UserObject` instance that
                             owns the property. Cannot be set in conjunction
                             with `default`.
    """

    def __init__(
            self,
            property_type: Union[Type, Schema],
            name: Optional[str] = None,
            default: Optional[Any] = None,
            default_function: Optional[Callable] = None
    ) -> None:
        if default and default_function:
            raise ValueError("Cannot use both default and default_function.")
        self.schema_type = resolve_arg_to_schema(property_type)
        self.name = name
        self.default = default
        self.default_function = default_function

    def __get__(
            self,
            instance: DOMObject,
            owner: Type[DOMObject]
    ) -> Any:
        if instance is None:
            raise AttributeError(
                "JSONSchemaProperty is not valid as a class descriptor")
        if self.name not in instance:
            if self.default_function:
                instance[self.name] = self.default_function(instance)
            else:
                instance[self.name] = self.default
        return instance[self.name]

    def __set__(
            self,
            instance: DOMObject,
            value: Any
    ) -> None:
        instance[self.name] = value



from typing import Dict, Union

from ..base_schema import Schema


class DOMProperties(object):

    properties: Dict[str, Schema] = None
    additional_properties: Union[bool, Schema] = False

    def __init__(
            self,
            properties: Dict[str, Schema] = None,
            additional_properties: Union[bool, Schema] = False
    ) -> None:
        self.properties = properties or {}
        self.additional_properties = additional_properties



from typing import Any, Dict, Tuple, Iterable

from ..dom import DOMInfo
from ..exceptions import ValidationError
from ..base_schema import Schema


class SchemaAnyOf(Schema):
    allowed_schemas: Tuple[Schema] = None

    def __init__(
            self,
            allowed_schemas: Iterable[Schema]
    ) -> None:
        self.allowed_schemas = tuple(allowed_schemas)

    def __call__(
            self,
            value: Any,
            dom_info: DOMInfo = None
    ) -> Any:
        valid_schemas = [
            allowed_schema
            for allowed_schema in self.allowed_schemas
            if allowed_schema.is_valid(value)
        ]
        if len(valid_schemas) > 1:
            raise ValidationError(
                "Ambiguous validation, more than one schema "
                f"is valid: {valid_schemas}"
            )
        if len(valid_schemas) == 0:
            raise ValidationError(
                "No valid schema was found for the supplied value"
            )
        return valid_schemas[0](value, dom_info)

    @property
    def schema(self) -> Dict[str, Any]:
        return {
            'anyOf': [
                allowed_schema.schema
                for allowed_schema in self.allowed_schemas
            ]
        }



from __future__ import annotations

from typing import Any, Iterator, NamedTuple, Optional

from abc import ABC, abstractmethod

from ..base_schema import Schema, SchemaAnything


class DOMInfo(NamedTuple):

    element: Optional[DOMElement] = None
    document: Optional[DOMElement] = None
    parent: Optional[DOMElement] = None
    element_key: Optional[str] = None


class DOMElement(ABC):

    __json_dom_info__: DOMInfo = None

    @abstractmethod
    def __init__(
            self,
            value: Any = None,
            json_dom_info: DOMInfo = None,
            **kwargs: Any
    ) -> None:
        if value is not None:
            raise ValueError(
                "The parameter 'value' must be handled by a non-abstract subclass."
            )
        if json_dom_info:
            self.__json_dom_info__ = DOMInfo(
                element=self,
                document=(
                    self if json_dom_info.document is None
                    else json_dom_info.document
                ),
                parent=json_dom_info.parent,
                element_key=json_dom_info.element_key
            )
        else:
            self.__json_dom_info__ = DOMInfo(
                element=self, document=self)

    @classmethod
    def __json_schema__(cls) -> Schema:
        return SchemaAnything()

    @abstractmethod
    def to_builtin(self) -> Any:
        pass

    def walk_elements(self) -> Iterator[DOMInfo]:
        yield self.__json_dom_info__



from typing import Any, Dict, Tuple

from .Schema import Schema


class SchemaAnything(Schema):

    def __call__(
            self,
            value: Any,
            dom_info: Tuple = None
    ) -> Any:
        return value

    @property
    def schema(self) -> Dict[str, Any]:
        return {}



from typing import Any, Tuple

from .SchemaType import SchemaType


class SchemaNone(SchemaType):

    type_name: str = 'null'

    def __call__(
            self,
            value: None,
            dom_info: Tuple = None
    ) -> Any:
        if value is not None:
            raise ValueError('Value can only be None.')
        return None
